import { context } from './context';
import { defaultConcurrentTimeout } from './constants';
export const defaultSuite = suite('');
function getCurrentSuite() {
    return context.currentSuite || defaultSuite;
}
function createSuiteCollector(name, factory = () => { }, mode) {
    const queue = [];
    const factoryQueue = [];
    const suiteBase = {
        name,
        mode,
        hooks: {
            beforeAll: [],
            afterAll: [],
            beforeEach: [],
            afterEach: [],
        },
    };
    const test = createTestCollector((name, fn, mode, concurrent) => {
        queue.push({
            name,
            mode,
            concurrent,
            suite: {},
            state: (mode !== 'run' && mode !== 'only') ? mode : undefined,
            fn,
        });
    });
    const collector = {
        name,
        mode,
        test,
        collect,
        clear,
        on: addHook,
    };
    function addHook(name, ...fn) {
        suiteBase.hooks[name].push(...fn);
    }
    function clear() {
        queue.length = 0;
        factoryQueue.length = 0;
    }
    async function collect(file) {
        factoryQueue.length = 0;
        if (factory)
            await factory(test);
        const tasks = [...factoryQueue, ...queue];
        const suite = Object.assign(Object.assign({}, suiteBase), { tasks,
            file });
        tasks.forEach((task) => {
            task.suite = suite;
            if (file)
                task.file = file;
        });
        return suite;
    }
    context.currentSuite = collector;
    context.suites.push(collector);
    return collector;
}
function createConcurrentOptions(timeout) {
    return { timeout: timeout !== null && timeout !== void 0 ? timeout : defaultConcurrentTimeout };
}
function createTestCollector(collectTask) {
    function test(name, fn) {
        collectTask(name, fn, 'run');
    }
    test.concurrent = concurrent;
    test.skip = skip;
    test.only = only;
    test.todo = todo;
    function concurrent(name, fn, timeout) {
        collectTask(name, fn, 'run', createConcurrentOptions(timeout));
    }
    concurrent.skip = (name, fn, timeout) => collectTask(name, fn, 'skip', createConcurrentOptions(timeout));
    concurrent.only = (name, fn, timeout) => collectTask(name, fn, 'only', createConcurrentOptions(timeout));
    concurrent.todo = todo;
    function skip(name, fn) {
        collectTask(name, fn, 'skip');
    }
    skip.concurrent = concurrent.skip;
    function only(name, fn) {
        collectTask(name, fn, 'only');
    }
    only.concurrent = concurrent.only;
    function todo(name) {
        collectTask(name, () => { }, 'todo');
    }
    todo.concurrent = todo;
    return test;
}
// apis
export const test = (function () {
    function test(name, fn) {
        return getCurrentSuite().test(name, fn);
    }
    function concurrent(name, fn, timeout) {
        return getCurrentSuite().test.concurrent(name, fn, timeout);
    }
    concurrent.skip = (name, fn, timeout) => getCurrentSuite().test.concurrent.skip(name, fn, timeout);
    concurrent.only = (name, fn, timeout) => getCurrentSuite().test.concurrent.only(name, fn, timeout);
    concurrent.todo = (name) => getCurrentSuite().test.concurrent.todo(name);
    function skip(name, fn) {
        return getCurrentSuite().test.skip(name, fn);
    }
    skip.concurrent = (name, fn, timeout) => getCurrentSuite().test.skip.concurrent(name, fn, timeout);
    function only(name, fn) {
        return getCurrentSuite().test.only(name, fn);
    }
    only.concurrent = (name, fn, timeout) => getCurrentSuite().test.only.concurrent(name, fn, timeout);
    function todo(name) {
        return getCurrentSuite().test.todo(name);
    }
    todo.concurrent = (name) => getCurrentSuite().test.todo.concurrent(name);
    test.concurrent = concurrent;
    test.skip = skip;
    test.only = only;
    test.todo = todo;
    return test;
})();
export function suite(suiteName, factory) {
    return createSuiteCollector(suiteName, factory, 'run');
}
suite.skip = (suiteName, factory) => createSuiteCollector(suiteName, factory, 'skip');
suite.only = (suiteName, factory) => createSuiteCollector(suiteName, factory, 'only');
suite.todo = (suiteName) => createSuiteCollector(suiteName, undefined, 'todo');
// alias
export const describe = suite;
export const it = test;
// hooks
export const beforeAll = (fn) => getCurrentSuite().on('beforeAll', fn);
export const afterAll = (fn) => getCurrentSuite().on('afterAll', fn);
export const beforeEach = (fn) => getCurrentSuite().on('beforeEach', fn);
export const afterEach = (fn) => getCurrentSuite().on('afterEach', fn);
// utils
export function clearContext() {
    context.suites.length = 0;
    defaultSuite.clear();
    context.currentSuite = defaultSuite;
}
